SystemOrganization addCategory: #'Collections-BitmapCharacterSet'!

Collection subclass: #BitmapCharacterSet
	instanceVariableNames: 'byteCharacters wideCharacters tally'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-BitmapCharacterSet'!

!BitmapCharacterSet commentStamp: 'monty 4/5/2015 23:53' prior: 0!
This class implements a set of Character objects similar to CharacterSet and WideCharacterSet, but it uses a bitmap internally to test if wide (multibyte) characters belong to it rather than using a Dictionary like WideCharacterSet does. For byte characters, a simple 256-element Array is used, the same as with CharacterSet, which is faster but uses more memory.

(used by XMLParser)!

!BitmapCharacterSet class methodsFor: 'inspecting' stamp: 'monty 4/19/2015 15:41'!
inspectorClass
	"does not use Set class>>inspectorClass because it is incompatible"
	^ Smalltalk tools inspectorClass! !

!BitmapCharacterSet class methodsFor: 'instance creation' stamp: 'JAAyer 1/19/2011 23:03'!
new
	^ self new: 256! !

!BitmapCharacterSet class methodsFor: 'instance creation' stamp: 'JAAyer 1/19/2011 23:04'!
new: aCapacity
	^ self basicNew initialize: aCapacity! !

!BitmapCharacterSet class methodsFor: 'instance creation' stamp: 'monty 4/5/2015 23:40'!
newFrom: aCollection
	"for portability"
	^ self new
		addAll: aCollection;
		yourself	! !

!BitmapCharacterSet methodsFor: 'comparing' stamp: 'JAA 3/31/2011 17:17'!
= anObject
	self == anObject
		ifTrue: [^ true].

	(self class == anObject class
		and: [tally = anObject size])
		ifFalse: [^ false].

	^ self allSatisfy: [:each | anObject includes: each].! !

!BitmapCharacterSet methodsFor: 'adding' stamp: 'monty 3/10/2016 12:12'!
add: aCharacter
	| asciiValue |

	"optimized for speed with inlining; do not refactor"
	(asciiValue := aCharacter asciiValue) < 256
		ifTrue: [
			(byteCharacters at: asciiValue + 1)
				ifFalse: [tally := tally + 1].
			byteCharacters
				at: asciiValue + 1
				put: true]
		ifFalse: [| byteIndex byte bitmask |
			"256 // 8 - 31 = 1 (first index), (256 + 8) // 8 - 31 = 2 (second), etc"
			byteIndex := asciiValue // 8 - 31.
			(wideCharacters isNil
				or: [byteIndex > wideCharacters size])
				ifTrue: [self growWideCharacterBitmapTo: (byteIndex * 1.5) asInteger].

			byte := wideCharacters at: byteIndex.
			bitmask := 16r80 bitShift: 0 - (asciiValue \\ 8).
			
			"Only increment the tally if the character is not already present"
			(byte bitAnd: bitmask) > 0
				ifFalse: [tally := tally + 1].

			wideCharacters
				at: byteIndex
				put: (byte bitOr: bitmask)].
	^ aCharacter.! !

!BitmapCharacterSet methodsFor: 'private' stamp: 'monty 4/5/2015 23:43'!
byteCharacters
	^ byteCharacters! !

!BitmapCharacterSet methodsFor: 'enumerating' stamp: 'monty 3/10/2016 14:10'!
byteCharactersDo: aBlock
	1 to: byteCharacters size do: [:i |
		(byteCharacters at: i)
			ifTrue: [aBlock value: (Character value: i - 1)]]! !

!BitmapCharacterSet methodsFor: 'accessing' stamp: 'Anonymous 11/12/2013 00:32'!
capacity
	^ byteCharacters size +
		(wideCharacters
			ifNil: [0]
			ifNotNil: [wideCharacters size * 8]).! !

!BitmapCharacterSet methodsFor: 'enumerating' stamp: 'monty 3/10/2016 12:22'!
do: aBlock
	self
		byteCharactersDo: aBlock;
		wideCharactersDo: aBlock! !

!BitmapCharacterSet methodsFor: 'private' stamp: 'monty 3/10/2016 10:28'!
growWideCharacterBitmapTo: aSizeInBytes
	| newSize |

	newSize := aSizeInBytes min: self maxWideCharactersSize.
	wideCharacters
		ifNil: [wideCharacters := ByteArray new: newSize]
		ifNotNil: [
			wideCharacters :=
				(ByteArray new: newSize)
					replaceFrom: 1
					to: wideCharacters size
					with: wideCharacters
					startingAt: 1].! !

!BitmapCharacterSet methodsFor: 'testing' stamp: 'monty 3/10/2016 12:24'!
hasWideCharacters
	self wideCharactersDo: [:each | ^ true].
	^ false.! !

!BitmapCharacterSet methodsFor: 'comparing' stamp: 'monty 3/10/2016 11:17'!
hash
	^ self class hash bitXor: self size hash! !

!BitmapCharacterSet methodsFor: 'testing' stamp: 'monty 3/10/2016 12:13'!
includes: aCharacter
	| asciiValue |

	"optimized for speed with inlining; do not refactor"
	(asciiValue := aCharacter asciiValue) < 256
		ifTrue: [^ byteCharacters at: asciiValue + 1]
		ifFalse: [| byteIndex |
			wideCharacters
				ifNil: [^ false].
			"256 // 8 - 31 = 1 (first index), (256 + 8) // 8 - 31 = 2 (second), etc"
			(byteIndex := asciiValue // 8 - 31) > wideCharacters size
				ifTrue: [^ false].
			^ ((wideCharacters at: byteIndex) bitAnd:
				(16r80 bitShift: 0 - (asciiValue \\ 8))) > 0]! !

!BitmapCharacterSet methodsFor: 'initialization' stamp: 'monty 3/10/2016 12:03'!
initialize: aCapacity
	byteCharacters := Array new: 256 withAll: false.
	aCapacity > 256
		ifTrue: [
			"(257 - 1) // 8 - 31 = 1 (first byte),
			(257 + 8 - 1) // 8 - 31 = 2 (second byte), etc"
			self growWideCharacterBitmapTo: (aCapacity - 1) // 8 - 31].
	tally := 0.! !

!BitmapCharacterSet methodsFor: 'private' stamp: 'monty 3/10/2016 10:27'!
maxWideCharactersSize
	^ 139232 "16r10FFFF // 8 - 31"! !

!BitmapCharacterSet methodsFor: 'copying' stamp: 'JAAyer 1/20/2011 01:05'!
postCopy
	byteCharacters := byteCharacters copy.
	wideCharacters
		ifNotNil: [wideCharacters := wideCharacters copy].! !

!BitmapCharacterSet methodsFor: 'removing' stamp: 'monty 3/10/2016 12:12'!
remove: aCharacter ifAbsent: aBlock
	| asciiValue |

	"optimized for speed with inlining; do not refactor"
	(asciiValue := aCharacter asciiValue) < 256
		ifTrue: [
			(byteCharacters at: asciiValue + 1)
				ifFalse: [^ aBlock value].
			byteCharacters
				at: asciiValue + 1
				put: false]
		ifFalse: [| byteIndex byte bitmask |
			"256 // 8 - 31 = 1 (first index), (256 + 8) // 8 - 31 = 2 (second), etc"
			byteIndex := asciiValue // 8 - 31.
			(wideCharacters isNil
				or: [byteIndex > wideCharacters size])
				ifTrue: [^ aBlock value].

			bitmask := 16r80 bitShift: 0 - (asciiValue \\ 8).
			((byte := wideCharacters at: byteIndex) bitAnd: bitmask) > 0
				ifFalse: [^ aBlock value].

			wideCharacters
				at: byteIndex
				put: (byte bitAnd: bitmask bitInvert)].
	tally := tally - 1.
	^ aCharacter.! !

!BitmapCharacterSet methodsFor: 'removing' stamp: 'monty 3/10/2016 14:12'!
removeAll
	"empties but preserves the capacity"

	1 to: byteCharacters size do: [:i |
		byteCharacters
			at: i
			put: false].
	wideCharacters
		ifNotNil: [
			1 to: wideCharacters size do: [:i |
				wideCharacters
					at: i
					put: 0]].
	tally := 0.! !

!BitmapCharacterSet methodsFor: 'accessing' stamp: 'JAAyer 1/21/2011 17:29'!
size
	^ tally! !

!BitmapCharacterSet methodsFor: 'private' stamp: 'monty 4/5/2015 23:43'!
wideCharacters
	^ wideCharacters! !

!BitmapCharacterSet methodsFor: 'enumerating' stamp: 'monty 3/10/2016 14:00'!
wideCharactersDo: aBlock
	"optimized for speed with to:do: and inlining; do not refactor"
	| baseValue |

	wideCharacters
		ifNil: [^ self].

	baseValue := 248. "256 - 8"
	1 to: wideCharacters size do: [:byteIndex | | byte |
		baseValue := baseValue + 8. "248 + 8 = 256, 256 + 8 = 264, etc"
		(byte := wideCharacters at: byteIndex) > 0
			ifTrue: [
				0 to: 7 do: [:shiftIndex |
					(byte bitAnd: (16r80 bitShift: 0 - shiftIndex)) > 0
						ifTrue: [
							aBlock value:
								(Character value: baseValue + shiftIndex)]]]].! !
